@@ -2,18 +2,17 @@ package Astro::FITS::CFITSIO::FileName;
 
 # ABSTRACT: parse and generate CFITSIO extended file names.
 
-## no critic (ProhibitSubroutinePrototypes)
+use v5.26;
 
 use strict;
 use warnings;
 
-use v5.26;
+our $VERSION = '0.06';
 
-our $VERSION = '0.05';
-
-use Types::Standard qw[ Str ArrayRef StrMatch Enum Bool Optional Dict ];
+use Types::Standard        qw[ Str ArrayRef StrMatch Enum Bool Optional Dict ];
 use Types::Common::Numeric qw[ PositiveOrZeroInt PositiveInt ];
-use Ref::Util qw[ is_arrayref ];
+use Scalar::Util           qw[ blessed ];
+use Ref::Util              qw[ is_arrayref ];
 
 use Astro::FITS::CFITSIO::FileName::Regexp -all;
 
@@ -28,13 +27,12 @@ use experimental 'signatures', 'postderef', 'declared_refs', 'refaliasing';
 
 use namespace::clean;
 
-use overload '""', "_stringify", fallback => 1;
-
-
-
-
-
+BEGIN { with 'MooX::Tag::TO_HASH'; }
 
+use overload
+  q{""}    => '_stringify',
+  bool     => sub { die "don't use ${\__PACKAGE__} in boolean context" },
+  fallback => 1;
 
 
 
@@ -181,6 +179,7 @@ has base_filename => (
     is       => 'ro',
     isa      => Str,
     required => 1,
+    to_hash  => 1,
 );
 
 
@@ -201,6 +200,7 @@ has file_type => (
     is        => 'ro',
     isa       => Str,
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -221,6 +221,7 @@ has output_name => (
     is        => 'ro',
     isa       => Str,
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -241,6 +242,7 @@ has extname => (
     is        => 'ro',
     isa       => Str,
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -261,6 +263,7 @@ has extver => (
     is        => 'ro',
     isa       => PositiveInt,
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -280,10 +283,10 @@ has extver => (
 
 
 has xtension => (
-    is => 'ro',
-    isa =>
-      StrMatch [qr{\A A | ASCII | I | IMAGE | T| TABLE | B | BINTABLE \z}xi],
+    is        => 'ro',
+    isa       => StrMatch [qr{\A [AITB] | ASCII | IMAGE | TABLE | BINTABLE \z}xi],
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -304,6 +307,7 @@ has image_cell_spec => (
     is        => 'ro',
     isa       => Str,
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -324,6 +328,7 @@ has hdunum => (
     is        => 'ro',
     isa       => PositiveOrZeroInt,
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -344,6 +349,7 @@ has compress_spec => (
     is        => 'ro',
     isa       => Str,
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -364,6 +370,7 @@ has image_section => (
     is        => 'ro',
     isa       => ArrayRef [ StrMatch [$PixelRange] ],
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -403,9 +410,10 @@ has pix_filter => (
     isa => Dict [
         datatype     => Optional [ Enum [qw( b i j r d )] ],
         discard_hdus => Optional [Bool],
-        expr         => Str
+        expr         => Str,
     ],
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -426,6 +434,7 @@ has col_filter => (
     is        => 'ro',
     isa       => ArrayRef [Str],
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -446,6 +455,7 @@ has row_filter => (
     is        => 'ro',
     isa       => ArrayRef [Str],
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -481,10 +491,11 @@ has bin_spec => (
     isa => ArrayRef [
         Dict [
             datatype => Optional [ Enum [qw( b i j r d )] ],
-            expr => Optional[Str]
-        ]
+            expr     => Optional [Str],
+        ],
     ],
     predicate => 1,
+    to_hash   => ',if_defined',
 );
 
 
@@ -496,7 +507,7 @@ has bin_spec => (
 has filename => (
     is       => 'lazy',
     init_arg => undef,
-    builder  => 1
+    builder  => 1,
 );
 
 
@@ -517,8 +528,7 @@ around BUILDARGS => sub ( $orig, $class, @args ) {
     my \%args = $class->$orig( @args );
 
     $args{$_} = [ $args{$_} ]
-      for grep { $args{$_} && !is_arrayref( $args{$_} ) }
-      qw( col_filter row_filter bin_spec );
+      for grep { $args{$_} && !is_arrayref( $args{$_} ) } qw( col_filter row_filter bin_spec );
 
     return \%args;
 };
@@ -527,21 +537,20 @@ sub BUILD ( $self, $ ) {
 
     my $is_image = $self->has_image_section || $self->has_pix_filter;
 
-    _croak(
-        "can't specify column, row, or bin/histogram specs if it's an image" )
+    _croak( q{can't specify column, row, or bin/histogram specs if it's an image} )
       if $is_image
       && ( $self->has_col_filter
         || $self->has_row_filter
         || $self->has_bin_spec );
 
-    _croak( "must specify extname if extver is specified" )
+    _croak( q{must specify extname if extver is specified} )
       if $self->has_extver && !$self->has_extname;
 
-    _croak( "don't specify both extname and hdunum" )
+    _croak( q{don't specify both extname and hdunum} )
       if $self->has_extname && $self->has_hdunum;
 
     # compress spec can only be used if nothing else is specified.
-    _croak( "specify compress spec by itself" )
+    _croak( q{specify compress spec by itself} )
       if $self->has_compress_spec
       && ( $self->has_extname
         || $self->has_extver
@@ -564,9 +573,11 @@ sub BUILD ( $self, $ ) {
 
 
 
+
+
 sub parse_filename ( $, $filename ) {
 
-    _croak( "can't parse filename: @{[ $filename // '<undef>' ]}" )
+    _croak( q|can't parse filename: @{[ $filename // '<undef>' ]}| )
       unless defined $filename && $filename =~ $FileName;
 
     my %match = %+;
@@ -574,40 +585,31 @@ sub parse_filename ( $, $filename ) {
     if ( exists $match{col_bin_row} ) {
         # * didn't match [image_section][pix filter]
         # * found a bunch of things.
-        for my $spec (
-            $match{col_bin_row} =~ /( \[ $PossiblyQuotedStringInSpec \] )/xg )
-        {
+        for my $spec ( $match{col_bin_row} =~ /( \[ $PossiblyQuotedStringInSpec \] )/xg ) {
 
             if ( $spec =~ $binSpec ) {
                 push(
                     ( $match{bin_spec} //= [] )->@*,
                     { (
-                            exists $+{bin_spec_expression}
-                            ? ( expr => $+{bin_spec_expression} )
+                            exists $+{bin_spec_expression} ? ( expr => $+{bin_spec_expression} )
                             : ()
                         ),
                         (
-                            exists $+{bin_spec_datatype}
-                            ? ( datatype => $+{bin_spec_datatype} )
+                            exists $+{bin_spec_datatype} ? ( datatype => $+{bin_spec_datatype} )
                             : (),
-                        )
+                        ),
                     },
                 );
             }
             elsif ( $spec =~ $colFilter ) {
 
-                push(
-                    ( $match{col_filter} //= [] )->@*,
-                    $+{col_filter} =~ m/$PossiblyQuotedStringInList/g
-                );
+                push( ( $match{col_filter} //= [] )->@*, $+{col_filter} =~ m/$PossiblyQuotedStringInList/g );
             }
             elsif ( $spec =~ $rowFilter ) {
                 push( ( $match{row_filter} //= [] )->@*, $+{row_filter} );
             }
             else {
-                _croak(
-                    "$spec is not a bin/histogram spec, column filter, or row filter"
-                );
+                _croak( q{spec is not a bin/histogram spec, column filter, or row filter} );
             }
         }
         delete $match{col_bin_row};
@@ -617,9 +619,8 @@ sub parse_filename ( $, $filename ) {
       = exists( $match{image_section_x} ) + exists( $match{image_section_y} );
 
     if ( $image_section ) {
-        _croak(
-            "internal error: missing either image_section_x or image_section_y"
-        ) unless $image_section == 2;
+        _croak( q{internal error: missing either image_section_x or image_section_y} )
+          unless $image_section == 2;
         $match{image_section}
           = [ delete @match{ 'image_section_x', 'image_section_y' } ];
     }
@@ -633,8 +634,7 @@ sub parse_filename ( $, $filename ) {
                 : ()
             ),
             (
-                exists $match{pix_filter_datatype}
-                ? ( datatype => delete $match{pix_filter_datatype} )
+                exists $match{pix_filter_datatype} ? ( datatype => delete $match{pix_filter_datatype} )
                 : ()
             ),
             expr => delete $match{pix_filter_expression},
@@ -643,7 +643,7 @@ sub parse_filename ( $, $filename ) {
     }
 
     # remove undefined entries
-    delete @match { grep ! defined $match{$_}, keys %match };
+    delete @match{ grep !defined $match{$_}, keys %match };
 
     \%match;
 }
@@ -685,19 +685,19 @@ sub render_base_filename ( $self ) {
 }
 
 sub render_file_type ( $self ) {
-    $self->has_file_type ? $self->file_type : '';
+    $self->has_file_type ? $self->file_type : q{};
 }
 
 sub render_output_name ( $self ) {
     $self->has_output_name
       ? sprintf( '(%s)', $self->output_name )
-      : '';
+      : q{};
 }
 
 sub render_compress_spec ( $self ) {
     return $self->has_compress_spec
       ? sprintf( '[compress %s]', $self->compress_spec )
-      : '';
+      : q{};
 }
 
 sub render_hdu ( $self ) {
@@ -705,92 +705,84 @@ sub render_hdu ( $self ) {
     my @render;
 
     if ( $self->has_hdunum ) {
-        push @render, '[', $self->hdunum;
-        push @render, ';', $self->image_cell_spec
+        push @render, q{[}, $self->hdunum;
+        push @render, q{;}, $self->image_cell_spec
           if $self->has_image_cell_spec;
-        push @render, ']';
+        push @render, q{]};
     }
 
     elsif ( $self->has_extname ) {
-        push @render, '[', $self->extname;
-        push @render, ',', $self->extver if $self->has_extver;
-        push @render, ',', $self->xtension if $self->has_xtension;
-        push @render, ';', $self->image_cell_spec
+        push @render, q{[}, $self->extname;
+        push @render, q{,}, $self->extver   if $self->has_extver;
+        push @render, q{,}, $self->xtension if $self->has_xtension;
+        push @render, q{;}, $self->image_cell_spec
           if $self->has_image_cell_spec;
-        push @render, ']';
+        push @render, q{]};
     }
 
-    return join ('', @render );
+    return join( q{}, @render );
 }
 
 sub render_image_section ( $self ) {
 
     $self->has_image_section
       ? sprintf( '[%s,%s]', $self->image_section->[0], $self->image_section->[1] )
-      : '';
+      : q{};
 }
 
 sub render_pix_filter ( $self ) {
 
-    return '' unless $self->has_pix_filter;
+    return q{} unless $self->has_pix_filter;
 
     my @render;
 
-    push @render, '[', 'pix';
+    push @render, q{[}, 'pix';
     push @render, $self->pix_filter->{datatype}
       if exists $self->pix_filter->{datatype};
-    push @render, '1'
+    push @render, q{1}
       if exists $self->pix_filter->{discard_hdus}
       && $self->pix_filter->{discard_hdus};
-    push @render, ' ';
+    push @render, q{ };
     push @render, $self->pix_filter->{expr}, ']';
 
-    return join( '', @render );
+    return join( q{}, @render );
 
 }
 
 sub render_col_filter ( $self ) {
     return $self->has_col_filter
-      ? '[col ' . join(';', $self->col_filter->@* ) . ']'
-      : '';
+      ? '[col ' . join( q{;}, $self->col_filter->@* ) . q{]}
+      : q{};
 }
 
 sub render_row_filter ( $self ) {
     return $self->has_row_filter
-      ? join('', map { "[$_]" } $self->row_filter->@* )
-      : '';
+      ? join( q{}, map { "[$_]" } $self->row_filter->@* )
+      : q{};
 }
 
 sub render_bin_spec ( $self ) {
 
-    return '' unless $self->has_bin_spec;
+    return q{} unless $self->has_bin_spec;
 
     my @render;
 
     for ( $self->bin_spec->@* ) {
         push @render, '[bin';
         push @render, $_->{datatype} if exists $_->{datatype};
-        push @render, ' ', $_->{expr} if exists $_->{expr};
-        push @render, ']';
+        push @render, q{ }, $_->{expr} if exists $_->{expr};
+        push @render, q{]};
     }
 
-    return join( '', @render );
+    return join( q{}, @render );
 }
 
 sub _build_filename ( $self ) {
-    return join(
-        '',
-        $self->render_file_type,
-        $self->render_base_filename,
-        $self->render_output_name,
-        $self->render_compress_spec,
-        $self->render_hdu,
-        $self->render_image_section,
-        $self->render_pix_filter,
-        $self->render_col_filter,
-        $self->render_row_filter,
-        $self->render_bin_spec,
-    );
+    return join( q{},
+        $self->render_file_type,     $self->render_base_filename, $self->render_output_name,
+        $self->render_compress_spec, $self->render_hdu,           $self->render_image_section,
+        $self->render_pix_filter,    $self->render_col_filter,    $self->render_row_filter,
+        $self->render_bin_spec, );
 }
 
 
@@ -802,35 +794,23 @@ sub _build_filename ( $self ) {
 
 
 
-sub _maybe ( $self, $attr ) {
 
-    my $mth = "has_$attr";
-    return ( $self->$mth ? ( $attr => $self->$attr ) : () );
+
+# backwards compatibility;
+sub to_hash;
+*to_hash = \&TO_HASH;
+
+
+
+
+
+
+
+sub clone_with ( $self, @args ) {
+    my %attr = @args == 1 ? $args[0]->%* : @args;
+    return blessed( $self )->new( ( $self->TO_HASH->%*, %attr ) );
 }
 
-sub to_hash ( $self ) {
-
-    return {
-        base_filename => $self->base_filename,
-        map { $self->_maybe( $_ ) }
-          qw( file_type
-          output_name
-          extname
-          extver
-          xtension
-          image_cell_spec
-          hdunum
-          compress_spec
-          image_section
-          pix_filter
-          col_filter
-          row_filter
-          bin_spec
-           )
-           };
-}
-
-
 1;
 
 #
@@ -856,7 +836,7 @@ Astro::FITS::CFITSIO::FileName - parse and generate CFITSIO extended file names.
 
 =head1 VERSION
 
-version 0.05
+version 0.06
 
 =head1 SYNOPSIS
 
@@ -896,7 +876,7 @@ set, use the C<has_pix_filter> method.
 
 =head2 Manipulating a filename
 
-L<Astro::FITS::CFITSIO::FileName> objects are meant to be immutable, so
+C<Astro::FITS::CFITSIO::FileName> objects are meant to be immutable, so
 manipulating attributes has to be done outside of the object.
 
 One way is to call the L</parse_filename> class method on the original
@@ -905,7 +885,132 @@ L</to_hash> object method, which returns a hash of attributes for a
 particular object.  Both of these hashes may be fed into the class
 constructor.
 
-=head1 CLASS METHODS
+=head1 OBJECT ATTRIBUTES
+
+=head2 base_filename
+
+The name of the file, without any trailing specifications
+(e.g. anything in square brackets or parenthesis) or a leading URI.
+
+This attribute is always present.
+
+=head2 file_type
+
+The type of file.  CFITSIO uses a URI style specification (e.g. C<mem://>, C<http://> ).
+
+This is optional.  See L</has_file_type>
+
+=head2 output_name
+
+An output or template name, depending upon the context that it's used with CFITSIO.
+
+This is optional.  See L</has_output_name>
+
+=head2 extname
+
+The name of the HDU provided by the C<EXTNAME> header keyword.
+
+This is optional. See L</has_extname>
+
+=head2 extver
+
+The version of the HDU specified by the L</extname> attribute.  L</extname> must also be specified.
+
+This is optional.  See L</has_extver>.
+
+=head2 xtension
+
+The type of the HDU.  It is case insensitive and may be one of
+
+ A | ASCII |  I | IMAGE | T | TABLE | B | BINTABLE
+
+This is optional.  See L</has_xtension>
+
+=head2 image_cell_spec
+
+Images can be stored in vector cells; this specifies which cell to access.
+
+This is optional. See L</has_image_cell_spec>.
+
+=head2 hdunum
+
+The HDU index. C<0> is the primary HDU.  Do not use this with L</extname>.
+
+This is optional. See L</has_hdunum>.
+
+=head2 compress_spec
+
+The image tile compression specification. The specification is not validated.
+
+This is optional. See L</has_compress_spec>.
+
+=head2 image_section
+
+Returns an arrayref of two elements containing (as strings) the pixel ranges for the image axes.
+
+This is optional. See L</has_image_section>.
+
+=head2 pix_filter
+
+A hashref containing the pixel filter.
+
+The hash has the following entries:
+
+=over
+
+=item datatype
+
+Optional.
+
+=item discard_hdus
+
+Optional.
+
+=item expr
+
+The filter expression.
+
+=back
+
+This is optional.  See L</has_pix_filter>.
+
+=head2 col_filter
+
+Returns an array of column filters, as strings.
+
+This is optional.  See L</has_col_filter>.
+
+=head2 row_filter
+
+Returns an array of row filters, as strings.
+
+This is optional. See L</has_row_filter>
+
+=head2 bin_spec
+
+Returns an arrayref of hashrefs containing binning/histogram specifications.
+
+The hashes have the following entries:
+
+=over
+
+=item datatype
+
+Optional.
+
+=item expr
+
+The binning expression.
+
+=back
+
+This is optional. See L</has_bin_spec>.
+
+=head2 filename
+
+This returns the full CFITSIO filename based on all of the attributes.
+
+=head1 CONSTRUCTORS
 
 =head2 new
 
@@ -1035,147 +1140,26 @@ The binning expression.
 
 =back
 
+=head2 clone_with
+
+  $new = $old->clone_with( %attr | \%attr );
+
+=head1 CLASS METHODS
+
 =head2 parse_filename
 
   $attr = $class->parse_filename( $filename );
 
-Parses an extended CFITSIO filename and returns a hash which may be fed
-into the L<Astro::FITS::CFITSIO::FileName> constructor.   Typically
-it's easier to just call the class constructor with the filename as a single argument,
+Parses an extended CFITSIO filename and returns a hash which may be
+fed into the C<Astro::FITS::CFITSIO::FileName> constructor.  Typically
+it's easier to just call the class constructor with the filename as a
+single argument.
 
-If a subclass needs to amend the parsing, this is the method to override.
-
-=head1 ATTRIBUTES
-
-=head2 base_filename
-
-The name of the file, without any trailing specifications
-(e.g. anything in square brackets or parenthesis) or a leading URI.
-
-This attribute is always present.
-
-=head2 file_type
-
-The type of file.  CFITSIO uses a URI style specification (e.g. C<mem://>, C<http://> ).
-
-This is optional.  See L</has_file_type>
-
-=head2 output_name
-
-An output or template name, depending upon the context that it's used with CFITSIO.
-
-This is optional.  See L</has_output_name>
-
-=head2 extname
-
-The name of the HDU provided by the C<EXTNAME> header keyword.
-
-This is optional. See L</has_extname>
-
-=head2 extver
-
-The version of the HDU specified by the L</extname> attribute.  L</extname> must also be specified.
-
-This is optional.  See L</has_extver>.
-
-=head2 xtension
-
-The type of the HDU.  It is case insensitive and may be one of
-
- A | ASCII |  I | IMAGE | T | TABLE | B | BINTABLE
-
-This is optional.  See L</has_xtension>
-
-=head2 image_cell_spec
-
-Images can be stored in vector cells; this specifies which cell to access.
-
-This is optional. See L</has_image_cell_spec>.
-
-=head2 hdunum
-
-The HDU index. C<0> is the primary HDU.  Do not use this with L</extname>.
-
-This is optional. See L</has_hdunum>.
-
-=head2 compress_spec
-
-The image tile compression specification. The specification is not validated.
-
-This is optional. See L</has_compress_spec>.
-
-=head2 image_section
-
-Returns an arrayref of two elements containing (as strings) the pixel ranges for the image axes.
-
-This is optional. See L</has_image_section>.
-
-=head2 pix_filter
-
-A hashref containing the pixel filter.
-
-The hash has the following entries:
-
-=over
-
-=item datatype
-
-Optional.
-
-=item discard_hdus
-
-Optional.
-
-=item expr
-
-The filter expression.
-
-=back
-
-This is optional.  See L</has_pix_filter>.
-
-=head2 col_filter
-
-Returns an array of column filters, as strings.
-
-This is optional.  See L</has_col_filter>.
-
-=head2 row_filter
-
-Returns an array of row filters, as strings.
-
-This is optional. See L</has_row_filter>
-
-=head2 bin_spec
-
-Returns an arrayref of hashrefs containing binning/histogram specifications.
-
-The hashes have the following entries:
-
-=over
-
-=item datatype
-
-Optional.
-
-=item expr
-
-The binning expression.
-
-=back
-
-This is optional. See L</has_bin_spec>.
-
-=head2 filename
-
-This returns the full CFITSIO filename based on all of the attributes.
+If a subclass needs to amend the parsing, this is the method to
+override.
 
 =head1 METHODS
 
-=head2 has_bin_spec
-
-returns true if the L</bin_spec> attribute is present.
-
 =head2 has_file_type
 
 returns true if the L</file_type> attribute is present.
@@ -1257,9 +1241,11 @@ might return
 
   [extname, 2]
 
-=head2 to_hash
+=head2 to_hash (deprecated)
 
-  $hash = $self->to_hash;
+=head2 TO_HASH
+
+  $hash = $self->TO_HASH;
 
 returns a hashref containing the object's attributes.  This can be used to modify the attributes
 and then pass them to the class constructor to create a modified object.
@@ -1270,6 +1256,8 @@ and then pass them to the class constructor to create a modified object.
 
 Stringification is overloaded to return the filename as returned by the L</filename> attribute.
 
+=head1 INTERNALS
+
 =for Pod::Coverage BUILD
 BUILDARGS
 
@@ -1277,7 +1265,7 @@ BUILDARGS
 
 =head2 Bugs
 
-Please report any bugs or feature requests to bug-astro-fits-cfitsio-filename@rt.cpan.org  or through the web interface at: https://rt.cpan.org/Public/Dist/Display.html?Name=Astro-FITS-CFITSIO-FileName
+Please report any bugs or feature requests to bug-astro-fits-cfitsio-filename@rt.cpan.org  or through the web interface at: L<https://rt.cpan.org/Public/Dist/Display.html?Name=Astro-FITS-CFITSIO-FileName>
 
 =head2 Source
 
